'''
pire

Python Interactive Regular Expressions
'''
__title__ = 'pire'
__version__ = '0.1.1'
__all__ = ()
__author__ = 'Johan Nestaas <johannestaas@gmail.com'
__license__ = 'GPLv3'
__copyright__ = 'Copyright 2019 Johan Nestaas'

from .ui import run_pire


def main():
    import os
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--regex', '-r', default='regex.pire',
        help='path to text file with newline-delimited regular expressions',
    )
    parser.add_argument(
        'input_paths', nargs='+', help='paths to files to parse with regex',
    )
    args = parser.parse_args()
    if not os.path.exists(args.regex):
        with open(args.regex, 'w') as f:
            f.write('')
    run_pire(
        regex_path=args.regex,
        input_paths=args.input_paths,
    )



import os

from .exceptions import EditorMissingError


def _split_path():
    return os.getenv('PATH').split(':')


def which(binary):
    for path_dir in _split_path():
        path = os.path.join(path_dir, binary)
        if os.path.isfile(path):
            return path
    return None


def get_editor():
    editor = os.getenv('EDITOR')
    if editor:
        if which(editor):
            return which(editor)
        raise EditorMissingError(f'cant find editor {editor!r}')
    for editor in ('nano', 'vim', 'vi', 'emacs'):
        binary = which(editor)
        if binary:
            return binary
    raise EditorMissingError('cant find any editor from: nano, vim, vi, emacs')



import os
from setuptools import setup

# pire
# Python Interactive Regular Expression


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="pire",
    version="0.1.1",
    description="Python Interactive Regular Expression",
    author="Johan Nestaas",
    author_email="johannestaas@gmail.com",
    license="GPLv3",
    keywords="python regex re regular expressions",
    url="https://github.com/johannestaas/pire",
    packages=['pire'],
    package_dir={'pire': 'pire'},
    long_description=read('README.rst'),
    classifiers=[
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Environment :: X11 Applications :: Qt',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
    ],
    install_requires=['ezcurses'],
    entry_points={
        'console_scripts': [
            'pire=pire:main',
        ],
    },
)



import re
from collections import namedtuple

from .exceptions import PireError


def _color_gen():
    colors = ('green', 'blue', 'magenta', 'cyan', 'red')
    color_i = 0
    while True:
        yield colors[color_i]
        color_i += 1
        color_i %= len(colors)


# Tracks where color text is inserted into lines for matched groups.
_Insert = namedtuple('_Insert', ('text', 'color'))


def _after_i_func(after):
    def func(i):
        return after + i + 1
    return func


def _before_i_func(before):
    def func(i):
        return before - i - 1
    return func


class Match:

    def __init__(self, line, match, names):
        self.line = line
        self.match = match
        self.names = names

    def draw(self, win, pos):
        # Return yellow if not a match.
        if self.match is None:
            win.write(self.line, color='yellow', pos=pos)
            return

        # Return default color if no groups defined.
        if not self.match.groups():
            win.write(self.line, color=None, pos=pos)
            return

        # Now we have to show the groups.
        self._draw_groups(win, pos)

    def _draw_groups(self, win, pos):
        # Sort by the match span, by start and longest ones.
        def _inverse_span_sort(name):
            span_start, span_end = self.match.span(name)
            # If they're equal starts, then take the longest one as the first.
            return (span_start, span_end * -1)
        names = sorted(self.names, key=_inverse_span_sort)
        inserts = []
        color_gen = _color_gen()
        for i, name in enumerate(names):
            if not self.match.group(name):
                continue
            color = next(color_gen)
            start, end = self.match.span(name)
            text = f'({name}:'
            inserts.append((start, i, _Insert(text, color)))
            inserts.append((end, -i, _Insert(')', color)))
        # This sorts it so it's naturally sorted by the order they should be
        # written in (i and -i make it so the later matches get written
        # inside).
        inserts = [
            (i_pos, ins)
            for i_pos, _, ins in sorted(inserts)
        ]

        x, y = pos
        last = 0
        for ipos, insert in inserts:
            if ipos > last:
                win.write(self.line[last:ipos], color=None, pos=(x, y))
                x += ipos - last
            win.write(insert.text, color=insert.color, pos=(x, y))
            x += len(insert.text)
            last = ipos
        if self.line[last:]:
            win.write(self.line[last:], color=None, pos=(x, y))

    def _match_is_none(self):
        return self.match is None

    def _match_is_not_none(self):
        return self.match is not None


class Regex:

    def __init__(self, regex_str):
        self.regex_str = regex_str
        self.regex = re.compile(regex_str)
        self.num_groups = self.regex.groups
        if self.regex.groupindex.keys():
            self.group_names = list(self.regex.groupindex.keys())
        elif self.num_groups:
            self.group_names = list(range(1, self.num_groups + 1))
        else:
            self.group_names = None
        self.output = []

    def match(self, line):
        match = self.regex.match(line)
        self.output.append(Match(line, match, self.group_names))

    def first(self, after=None, before=None, match=True):
        if after is not None:
            slc = slice(after + 1, None, 1)
            calc_result = _after_i_func(after)
        elif before is not None:
            slc = slice(before - 1, None, -1)
            calc_result = _before_i_func(before)
        else:
            raise PireError('requires `before` or `after` keywords')
        if match:
            check = Match._match_is_not_none
        else:
            check = Match._match_is_none
        for i, match in enumerate(self.output[slc]):
            if check(match):
                return calc_result(i)
        return None


def load_regexes(path):
    regex_strs = []
    with open(path) as f:
        for line in f:
            regex_strs.append(line.rstrip('\n'))
    return [
        Regex(regex_str)
        for regex_str in regex_strs
    ]


def run_regex(regex, line_gen):
    for line in line_gen:
        line = line.rstrip('\n')
        regex.match(line)



from subprocess import call

from ezcurses import Cursed

from .regex_engine import load_regexes, run_regex
from .local_os import get_editor


def edit_regex(path):
    editor = get_editor()
    call([editor, path])


def gen_lines(input_paths):
    for path in input_paths:
        with open(path) as f:
            yield from f


def display(
    regexes, sel_regex, scr, regex_win, out_win, start=0, lines=None,
):
    """
    Displays the input matched against the regexes in two windows.
    """
    regex_win.clear()
    out_win.clear()
    for i, regex in enumerate(regexes):
        color = None if regex is not sel_regex else 'green'
        regex_win.write(regex.regex_str, pos=(0, i), color=color)
    regex_win.refresh()
    out_w, h = out_win.max_size()
    out_win.write('=' * out_w, pos=(0, 0), color='blue')
    if sel_regex is not None:
        output = sel_regex.output[start:start + h]
        for i, out in enumerate(output):
            pos = (0, i + 1)
            out.draw(out_win, pos)
    else:
        for i, line in enumerate(lines):
            if i < start:
                continue
            if i >= start + h - 1:
                break
            out_win.write(line, pos=(0, i - start + 1), color='yellow')
    out_win.write(
        (
            '?:help q:quit  e:edit  w:regex_up  s:regex_down  r:out_up  '
            'f:out_down  [:page_up  ]:page_down'
        ),
        pos=(0, h - 1),
        color=('white', 'blue'),
    )
    out_win.refresh()


def run_pire(
    regex_path=None,
    input_paths=None,
    split_horiz=True,
):
    with Cursed() as scr:
        sel_index = 0
        output_start = 0
        while True:
            regexes = load_regexes(regex_path)
            if not regexes:
                sel_regex = None
            else:
                sel_regex = regexes[sel_index]
            layout = scr.new_layout()
            layout.add_row(2, [12])
            layout.add_row(10, [12])
            layout.draw()
            rows = list(layout)
            regex_win = rows[0][0]
            out_win = rows[1][0]
            line_gen = gen_lines(input_paths)
            if sel_regex is not None:
                run_regex(sel_regex, line_gen)
                output = sel_regex.output
            else:
                output = list(line_gen)
            scr.clear()
            scr.refresh()
            display(
                regexes, sel_regex, scr, regex_win, out_win,
                start=output_start, lines=output,
            )
            key = scr.getkey()
            if key == 'q':
                break
            elif key == 'e':
                edit_regex(regex_path)
            elif key == '?':
                draw_help(scr)
            elif key == 's':
                sel_index += 1
                sel_index %= len(regexes)
            elif key == 'w':
                sel_index -= 1
                sel_index %= len(regexes)
            elif key == 'r':
                output_start = max(output_start - 1, 0)
            elif key == 'f':
                output_start = min(output_start + 1, len(output))
            elif key == '[':
                output_start = max(output_start - 10, 0)
            elif key == ']':
                output_start = min(output_start + 10, len(output))
            elif key == 'g':
                output_start = 0
            elif key == 'G':
                output_start = len(output) - 1
            elif sel_regex and key == 'n':
                first = sel_regex.first(after=output_start, match=True)
                if first is not None:
                    output_start = first
            elif sel_regex and key == 'N':
                first = sel_regex.first(before=output_start, match=True)
                if first is not None:
                    output_start = first
            elif sel_regex and key == 'm':
                first = sel_regex.first(after=output_start, match=False)
                if first is not None:
                    output_start = first
            elif sel_regex and key == 'M':
                first = sel_regex.first(before=output_start, match=False)
                if first is not None:
                    output_start = first


def draw_help(scr):
    w, h = scr.max_size()
    scr.clear()
    for i, txt in enumerate([
        '?: display this help',
        'q: quit',
        'e: open editor ($EDITOR environment variable)',
        's: select next regex',
        'w: select previous regex',
        'r: go up one line in output',
        'f: go down one line in output',
        '[: go up 10 lines',
        ']: go down 10 lines',
        'g: go to top of output',
        'G: go to bottom of output',
        'n: find next regex match after cursor',
        'N: find next regex match before cursor',
        'm: find next regex mismatch after cursor',
        'M: find next regex mismatch before cursor',
    ]):
        scr.write(txt, pos=(0, i), color=('white', 'blue'))
    scr.refresh()
    scr.getkey()
